# This module is part of lshash and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

__title__ = 'sparselsh'
__author__ = 'Brandon Roberts (brandon@bxroberts.org)'
__license__ = 'MIT'
__version__ = '2.1.1'

from sparselsh.lsh import LSH
